/**
 * Functions to contruct and assist the user interfaces with information.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>

#include "ui.h"
#include "zen_ui.h"
#include "retrieve.h"
#include "threads.h"

/* Temporary space for status line. This will move and use mutex locks. */
static char ui_status_text[1024] = "";

/**
 * Used as thread function to retrieve a new page.
 *
 * @param argument An array of strings with the URL and referer URL.
 *
 * @return always NULL.
 */
static void *threaded_request_page(void *arguments)
{
  char *url, *referer;
  struct layout_part *new_page;
  
  url = ((char **)arguments)[0];
  referer = ((char **)arguments)[1];

  new_page = retrieve_page(url, referer);
  thread_give_response(THREAD_RESPONSE_NEW_PAGE, (void *)new_page);

  return NULL;
}

/**
 * User interface helper function, used by the interface to request 
 * a new page.
 *
 * @param url The URL for the new page, relative or absolute.
 * @param referer The full referer URL, or NULL if there is no referer.
 *
 * @return non-zero value if an error occurred.
 */
static int ui_functions_request_page(char *url, char *referer)
{
  static char *arguments[2];
  int ret;

  arguments[0] = url;
  arguments[1] = referer;

  /* First kill off all old page requests. */
  thread_kill(THREAD_TYPE_GET_PAGE);

  ret = thread_start(THREAD_TYPE_GET_PAGE, threaded_request_page, 
		     (void *)arguments);

  return ret;
}

/**
 * User interface helper function, used by the interface to check if
 * a newly requested page is ready for displaying.
 *
 * @return a pointer to a page information part, containing the page
 * @return or NULL if there was no page ready for display.
 */
static struct layout_part *ui_functions_poll_page(void)
{
  struct layout_part *new_page = NULL;
  
  if(thread_check_response(THREAD_RESPONSE_NEW_PAGE, (void **)&new_page)) {
    return new_page;
  }

  return NULL;
}

/**
 * User interface helper function, used by the interface to get the current
 * status text.
 *
 * @param status A pointer to an area to store the status text in.
 * @param max_length The maximum length of the status text to store.
 *
 * @return the actual number of characters in the status text. 
 */
static int ui_functions_get_status(char *status, int max_length)
{
  strncpy(status, ui_status_text, max_length);
  status[max_length] = '\0';
  ui_status_text[0] = '\0';

  return strlen(status);
}

/**
 * Set the status text, which can be gotten from the user interface later.
 * This is used by various parts of the whole program, to report the 
 * current status about what it is doing, and if something special happens.
 *
 * @param status The new status text to store.
 *
 * @return the length of the new status text.
 */
int ui_functions_set_status(char *status)
{
  strncpy(ui_status_text, status, 1023);
  ui_status_text[1023] = '\0';

  return strlen(ui_status_text);
}

/**
 * Initialize the functions that the generic interface can offer to the
 * user interfaces.
 *
 * @return a pointer to the struct containing pointers to the functions.
 */
struct zen_ui_functions *ui_functions_init(void)
{
  static struct zen_ui_functions functions;

  if(user_interface.ui_support.interaction) {
    functions.get_page = retrieve_page;
    functions.request_page = ui_functions_request_page;
    functions.poll_page = ui_functions_poll_page;
    functions.get_status = ui_functions_get_status;
  } else {
    functions.get_page = retrieve_page;
    functions.request_page = NULL;
    functions.poll_page = NULL;
    functions.get_status = ui_functions_get_status;
  }

  return &functions;
}
